/* zynq.c
 *
 * Copyright (C) 2024 wolfSSL Inc.
 *
 * This file is part of wolfBoot.
 *
 * wolfBoot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * wolfBoot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335, USA
 */

#ifdef TARGET_zynq

#include "hal/zynq.h"

#ifndef ARCH_AARCH64
#   error "wolfBoot zynq HAL: wrong architecture selected. Please compile with ARCH=AARCH64."
#endif

#if defined(__QNXNTO__) && !defined(NO_QNX)
    #define USE_QNX
#elif defined(USE_BUILTIN_STARTUP)
    /* to use the Xilinx QSPI driver define USE_XQSPIPSU */
#endif

#include <target.h>
#include "image.h"
#include "printf.h"

#include <stdint.h>
#include <string.h>

#ifdef USE_XQSPIPSU
    /* Xilinx BSP Driver */
    #include "xqspipsu.h"
    #ifndef QSPI_DEVICE_ID
    #define QSPI_DEVICE_ID		XPAR_XQSPIPSU_0_DEVICE_ID
    #endif
    #ifndef QSPI_CLK_PRESACALE
    #define QSPI_CLK_PRESACALE  XQSPIPSU_CLK_PRESCALE_8
    #endif
#elif defined(USE_QNX)
    /* QNX QSPI driver */
    #include <sys/siginfo.h>
    #include "xzynq_gqspi.h"
#else
    /* QSPI bare-metal */
#endif

/* QSPI Slave Device Information */
typedef struct QspiDev {
    uint32_t mode;   /* GQSPI_GEN_FIFO_MODE_SPI, GQSPI_GEN_FIFO_MODE_DSPI or GQSPI_GEN_FIFO_MODE_QSPI */
    uint32_t bus;    /* GQSPI_GEN_FIFO_BUS_LOW, GQSPI_GEN_FIFO_BUS_UP or GQSPI_GEN_FIFO_BUS_BOTH */
    uint32_t cs;     /* GQSPI_GEN_FIFO_CS_LOWER, GQSPI_GEN_FIFO_CS_UPPER */
    uint32_t stripe; /* OFF=0 or ON=GQSPI_GEN_FIFO_STRIPE */
#ifdef USE_XQSPIPSU
    XQspiPsu qspiPsuInst;
#elif defined(USE_QNX)
    xzynq_qspi_t* qnx;
#endif
} QspiDev_t;

static QspiDev_t mDev;

/* forward declarations */
static int qspi_wait_ready(QspiDev_t* dev);
static int qspi_status(QspiDev_t* dev, uint8_t* status);
static int qspi_wait_we(QspiDev_t* dev);
#ifdef TEST_EXT_FLASH
static int test_ext_flash(QspiDev_t* dev);
#endif

#ifdef DEBUG_UART
void uart_init(void)
{
    /* Disable Interrupts */
    ZYNQMP_UART_IDR = ZYNQMP_UART_ISR_MASK;
    /* Disable TX/RX */
    ZYNQMP_UART_CR = (ZYNQMP_UART_CR_TX_DIS | ZYNQMP_UART_CR_RX_DIS);
    /* Clear ISR */
    ZYNQMP_UART_ISR = ZYNQMP_UART_ISR_MASK;

    /* 8-bits, no parity */
    ZYNQMP_UART_MR = ZYNQMP_UART_MR_PARITY_NONE;

    /* FIFO Trigger Level */
    ZYNQMP_UART_RXWM = 32; /* half of 64 byte FIFO */
    ZYNQMP_UART_TXWM = 32; /* half of 64 byte FIFO */

    /* RX Timeout - disable */
    ZYNQMP_UART_RXTOUT = 0;

    /* baud (115200) = master clk / (BR_GEN * (BR_DIV + 1)) */
    ZYNQMP_UART_BR_GEN = UART_MASTER_CLOCK / (DEBUG_UART_BAUD * (DEBUG_UART_DIV+1));
    ZYNQMP_UART_BR_DIV = DEBUG_UART_DIV;

    /* Reset TX/RX */
    ZYNQMP_UART_CR = (ZYNQMP_UART_CR_TXRST | ZYNQMP_UART_CR_RXRST);

    /* Enable TX/RX */
    ZYNQMP_UART_CR = (ZYNQMP_UART_CR_TX_EN | ZYNQMP_UART_CR_RX_EN);
}

void uart_write(const char* buf, uint32_t sz)
{
    uint32_t pos = 0;
    while (sz-- > 0) {
        char c = buf[pos++];
        if (c == '\n') { /* handle CRLF */
            while (ZYNQMP_UART_SR & ZYNQMP_UART_SR_TXFULL);
            ZYNQMP_UART_FIFO = '\r';
        }
        while (ZYNQMP_UART_SR & ZYNQMP_UART_SR_TXFULL);
        ZYNQMP_UART_FIFO = c;
    }
    /* Wait till TX Fifo is empty */
    while (!(ZYNQMP_UART_SR & ZYNQMP_UART_SR_TXEMPTY));
}
#endif /* DEBUG_UART */


#ifdef USE_XQSPIPSU
/* Xilinx BSP Driver */

/* Aligned page data buffer for DMA */
#ifdef __ICCARM__
#pragma data_alignment = 32
static uint8_t pageData[FLASH_PAGE_SIZE];
#pragma data_alignment = 4
#else
static uint8_t pageData[FLASH_PAGE_SIZE] __attribute__ ((aligned(32)));;
#endif
static int qspi_transfer(QspiDev_t* pDev,
    const uint8_t* cmdData, uint32_t cmdSz,
    const uint8_t* txData, uint32_t txSz,
    uint8_t* rxData, uint32_t rxSz, uint32_t dummySz,
    uint32_t mode)
{
    int ret;
    XQspiPsu_Msg msgs[4];
    uint32_t msgCnt = 0, busWidth = XQSPIPSU_SELECT_MODE_SPI;
    uint8_t* rxPtr = rxData;

    /* Chip Select */
    if (pDev->cs == GQSPI_GEN_FIFO_CS_BOTH) {
        XQspiPsu_SelectFlash(&pDev->qspiPsuInst,
            XQSPIPSU_SELECT_FLASH_CS_BOTH, XQSPIPSU_SELECT_FLASH_BUS_BOTH);
    }
    else if (pDev->cs == GQSPI_GEN_FIFO_CS_LOWER) {
        XQspiPsu_SelectFlash(&pDev->qspiPsuInst,
            XQSPIPSU_SELECT_FLASH_CS_LOWER, XQSPIPSU_SELECT_FLASH_BUS_LOWER);
    }
    else {
        XQspiPsu_SelectFlash(&pDev->qspiPsuInst,
            XQSPIPSU_SELECT_FLASH_CS_UPPER, XQSPIPSU_SELECT_FLASH_BUS_UPPER);
    }

    /* Transfer Bus Width - only applies to read/write command */
    if (mode == GQSPI_GEN_FIFO_MODE_QSPI)
        busWidth = XQSPIPSU_SELECT_MODE_QUADSPI;
    else if (mode == GQSPI_GEN_FIFO_MODE_DSPI)
        busWidth = XQSPIPSU_SELECT_MODE_DUALSPI;

    /* Command */
    memset(&msgs[msgCnt], 0, sizeof(XQspiPsu_Msg));
    msgs[msgCnt].TxBfrPtr = (uint8_t*)cmdData;
    msgs[msgCnt].ByteCount = cmdSz;
    msgs[msgCnt].BusWidth = XQSPIPSU_SELECT_MODE_SPI;
    msgs[msgCnt].Flags = XQSPIPSU_MSG_FLAG_TX;
    msgCnt++;

    /* TX */
    if (txData) {
        memset(&msgs[msgCnt], 0, sizeof(XQspiPsu_Msg));
        msgs[msgCnt].TxBfrPtr = (uint8_t*)txData;
        msgs[msgCnt].ByteCount = txSz;
        msgs[msgCnt].BusWidth = busWidth;
        msgs[msgCnt].Flags = XQSPIPSU_MSG_FLAG_TX;
        if (pDev->stripe & GQSPI_GEN_FIFO_STRIPE)
            msgs[msgCnt].Flags |= XQSPIPSU_MSG_FLAG_STRIPE;
        msgCnt++;
    }

    /* Dummy */
    if (dummySz > 0) {
        memset(&msgs[msgCnt], 0, sizeof(XQspiPsu_Msg));
        msgs[msgCnt].ByteCount = dummySz;
        msgs[msgCnt].BusWidth = busWidth;
        msgCnt++;
    }

    /* RX */
    if (rxData) {
        /* If RX pointer is not 32 byte aligned then use temp page data buffer */
        if (((size_t)rxPtr % 32) != 0)
            rxPtr = pageData;
        if (rxSz > (uint32_t)sizeof(pageData))
            rxSz = (uint32_t)sizeof(pageData);
        memset(&msgs[msgCnt], 0, sizeof(XQspiPsu_Msg));
        msgs[msgCnt].RxBfrPtr = rxPtr;
        msgs[msgCnt].ByteCount = rxSz;
        msgs[msgCnt].BusWidth = busWidth;
        msgs[msgCnt].Flags = XQSPIPSU_MSG_FLAG_RX;
        if (pDev->stripe & GQSPI_GEN_FIFO_STRIPE)
            msgs[msgCnt].Flags |= XQSPIPSU_MSG_FLAG_STRIPE;
        msgCnt++;
    }

    ret = XQspiPsu_PolledTransfer(&pDev->qspiPsuInst, msgs, msgCnt);
    if (ret < 0) {
        wolfBoot_printf("QSPI Transfer failed! %d\n", ret);
        return GQSPI_CODE_FAILED;
    }

    /* if unaligned read, return results */
    if (rxData && rxPtr == pageData) {
        memcpy(rxData, pageData, rxSz);
    }

    return GQSPI_CODE_SUCCESS;
}

#elif defined(USE_QNX)
/* QNX QSPI driver */
static int qspi_transfer(QspiDev_t* pDev,
    const uint8_t* cmdData, uint32_t cmdSz,
    const uint8_t* txData, uint32_t txSz,
    uint8_t* rxData, uint32_t rxSz, uint32_t dummySz,
    uint32_t mode)
{
    int ret;
    qspi_buf cmd_buf;
    qspi_buf tx_buf;
    qspi_buf rx_buf;
    uint32_t flags;

    flags = TRANSFER_FLAG_DEBUG;
    if (mode == GQSPI_GEN_FIFO_MODE_QSPI)
        flags |= TRANSFER_FLAG_MODE(TRANSFER_FLAG_MODE_QSPI);
    else if (mode == GQSPI_GEN_FIFO_MODE_DSPI)
        flags |= TRANSFER_FLAG_MODE(TRANSFER_FLAG_MODE_DSPI);
    else
        flags |= TRANSFER_FLAG_MODE(TRANSFER_FLAG_MODE_SPI);
    if (pDev->stripe & GQSPI_GEN_FIFO_STRIPE)
        flags |= TRANSFER_FLAG_STRIPE;
    if (pDev->cs & GQSPI_GEN_FIFO_CS_LOWER)
        flags |= TRANSFER_FLAG_LOW_DB | TRANSFER_FLAG_CS(TRANSFER_FLAG_CS_LOW);
    if (pDev->cs & GQSPI_GEN_FIFO_CS_UPPER)
        flags |= TRANSFER_FLAG_UP_DB | TRANSFER_FLAG_CS(TRANSFER_FLAG_CS_UP);

    memset(&cmd_buf, 0, sizeof(cmd_buf));
    cmd_buf.offset = (uint8_t*)cmdData;
    cmd_buf.len = cmdSz;

    memset(&tx_buf, 0, sizeof(tx_buf));
    tx_buf.offset = (uint8_t*)txData;
    tx_buf.len = txSz;

    memset(&rx_buf, 0, sizeof(rx_buf));
    rx_buf.offset = rxData;
    rx_buf.len = rxSz;

    /* Send the TX buffer */
    ret = xzynq_qspi_transfer(pDev->qnx,
        txData ? &tx_buf : NULL,
        rxData ? &rx_buf : NULL,
        &cmd_buf, flags);
    if (ret < 0) {
        wolfBoot_printf("QSPI Transfer failed! %d\n", ret);
        return GQSPI_CODE_FAILED;
    }
    return GQSPI_CODE_SUCCESS;
}
#else
/* QSPI bare-metal driver */
static inline int qspi_isr_wait(uint32_t wait_mask, uint32_t wait_val)
{
    uint32_t timeout = 0;
    while ((GQSPI_ISR & wait_mask) == wait_val &&
           ++timeout < GQSPI_TIMEOUT_TRIES);
    if (timeout == GQSPI_TIMEOUT_TRIES) {
        return -1;
    }
    return 0;
}

static int qspi_gen_fifo_write(uint32_t reg_genfifo)
{
    /* wait until the gen FIFO is not full to write */
    if (qspi_isr_wait(GQSPI_IXR_GEN_FIFO_NOT_FULL, 0)) {
        return GQSPI_CODE_TIMEOUT;
    }

#if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 3
    wolfBoot_printf("FifoEntry=%08x\n", reg_genfifo);
#endif
    GQSPI_GEN_FIFO = reg_genfifo;
    return GQSPI_CODE_SUCCESS;
}

static int gspi_fifo_tx(const uint8_t* data, uint32_t sz)
{
    uint32_t tmp32;
    while (sz > 0) {
        /* Wait for TX FIFO not full */
        if (qspi_isr_wait(GQSPI_IXR_TX_FIFO_FULL, GQSPI_IXR_TX_FIFO_FULL)) {
            return GQSPI_CODE_TIMEOUT;
        }

        /* Write data */
        if (sz >= 4) {
            GQSPI_TXD = *(uint32_t*)data;
            data += 4;
            sz -= 4;
        }
        else {
            tmp32 = 0;
            memcpy(&tmp32, data, sz);
            GQSPI_TXD = tmp32;
            sz = 0;
        }
    }
    return GQSPI_CODE_SUCCESS;
}

static int gspi_fifo_rx(uint8_t* data, uint32_t sz, uint32_t discardSz)
{
    uint32_t tmp32;
    while (sz > 0) {
        /* Wait for RX FIFO not empty */
        if (qspi_isr_wait(GQSPI_IXR_RX_FIFO_NOT_EMPTY, 0)) {
            return GQSPI_CODE_TIMEOUT;
        }
        if (discardSz >= GQSPI_FIFO_WORD_SZ) {
            tmp32 = GQSPI_RXD; /* discard */
            discardSz -= GQSPI_FIFO_WORD_SZ;
            continue;
        }
        if (sz >= 4) {
            *(uint32_t*)data = GQSPI_RXD;
            data += 4;
            sz -= 4;
        }
        else {
            tmp32 = GQSPI_RXD;
            memcpy(data, &tmp32, sz);
            sz = 0;
        }
    }
    return GQSPI_CODE_SUCCESS;
}

static int qspi_cs(QspiDev_t* pDev, int csAssert)
{
    uint32_t reg_genfifo;

    /* Select slave bus, bank, mode and cs clocks */
    reg_genfifo = (pDev->bus & GQSPI_GEN_FIFO_BUS_MASK);
    reg_genfifo |= GQSPI_GEN_FIFO_MODE_SPI;
    if (csAssert) {
        reg_genfifo |= (pDev->cs & GQSPI_GEN_FIFO_CS_MASK);
    }
    reg_genfifo |= GQSPI_GEN_FIFO_IMM(GQSPI_CS_ASSERT_CLOCKS);
    return qspi_gen_fifo_write(reg_genfifo);
}

static int qspi_transfer(QspiDev_t* pDev,
    const uint8_t* cmdData, uint32_t cmdSz,
    const uint8_t* txData, uint32_t txSz,
    uint8_t* rxData, uint32_t rxSz, uint32_t dummySz,
    uint32_t mode)
{
    int ret = GQSPI_CODE_SUCCESS;
    uint32_t reg_genfifo, xferSz;

    GQSPI_EN = 1; /* Enable device */
    qspi_cs(pDev, 1); /* Select slave */

    /* Setup bus slave selection */
    reg_genfifo = ((pDev->bus & GQSPI_GEN_FIFO_BUS_MASK) |
                   (pDev->cs & GQSPI_GEN_FIFO_CS_MASK) |
                    GQSPI_GEN_FIFO_MODE_SPI);

    /* Cmd Data */
    xferSz = cmdSz;
    while (ret == GQSPI_CODE_SUCCESS && cmdData && xferSz > 0) {
       /* Enable TX and send command inline */
       reg_genfifo |= GQSPI_GEN_FIFO_TX;
       reg_genfifo &= ~(GQSPI_GEN_FIFO_RX | GQSPI_GEN_FIFO_IMM_MASK);
       reg_genfifo |= GQSPI_GEN_FIFO_IMM(*cmdData); /* IMM is data */

       /* Submit general FIFO operation */
       ret = qspi_gen_fifo_write(reg_genfifo);
       if (ret != GQSPI_CODE_SUCCESS) {
           wolfBoot_printf("on line %d: error %d\n", __LINE__, ret);
           break;
       }

       /* offset size and buffer */
       xferSz--;
       cmdData++;
    }

    /* Set desired data mode and stripe */
    reg_genfifo |= (mode & GQSPI_GEN_FIFO_MODE_MASK);
    reg_genfifo |= (pDev->stripe & GQSPI_GEN_FIFO_STRIPE);

    /* TX Data */
    while (ret == GQSPI_CODE_SUCCESS && txData && txSz > 0) {
        xferSz = txSz;

        /* Enable TX */
        reg_genfifo &= ~(GQSPI_GEN_FIFO_RX | GQSPI_GEN_FIFO_IMM_MASK |
                         GQSPI_GEN_FIFO_EXP_MASK);
        reg_genfifo |= (GQSPI_GEN_FIFO_TX | GQSPI_GEN_FIFO_DATA_XFER);

        if (xferSz > GQSPI_GEN_FIFO_IMM_MASK) {
            /* Use exponent mode */
            xferSz = 256; /* 2 ^ 8 = 256 */
            reg_genfifo |= GQSPI_GEN_FIFO_EXP_MASK;
            reg_genfifo |= GQSPI_GEN_FIFO_IMM(8); /* IMM is exponent */
        }
        else {
            reg_genfifo |= GQSPI_GEN_FIFO_IMM(xferSz); /* IMM is length */
        }

        /* Submit general FIFO operation */
        ret = qspi_gen_fifo_write(reg_genfifo);
        if (ret != GQSPI_CODE_SUCCESS) {
            wolfBoot_printf("on line %d: error %d\n", __LINE__, ret);
        }

        /* Fill FIFO */
        ret = gspi_fifo_tx(txData, xferSz);
        if (ret != GQSPI_CODE_SUCCESS) {
            wolfBoot_printf("on line %d: error %d\n", __LINE__, ret);
            break;
        }

        /* offset size and buffer */
        txSz -= xferSz;
        txData += xferSz;
    }

    /* Dummy operations */
    if (ret == GQSPI_CODE_SUCCESS && dummySz) {
        /* Send dummy clocks (Disable TX & RX) */
        reg_genfifo &= ~(GQSPI_GEN_FIFO_TX | GQSPI_GEN_FIFO_RX |
                         GQSPI_GEN_FIFO_IMM_MASK | GQSPI_GEN_FIFO_EXP_MASK);
        /* IMM is number of dummy clock cycles */
        reg_genfifo |= GQSPI_GEN_FIFO_IMM(dummySz);
        ret = qspi_gen_fifo_write(reg_genfifo); /* Submit FIFO Dummy Op */

        if (rxSz > 0) {
            /* Convert dummy bits to bytes */
            dummySz = (dummySz + 7) / 8;
            /* Adjust rxSz for dummy bytes */
            rxSz += dummySz;
            /* round up by FIFO Word Size */
            rxSz = (((rxSz + GQSPI_FIFO_WORD_SZ - 1) / GQSPI_FIFO_WORD_SZ) *
                    GQSPI_FIFO_WORD_SZ);
        }
    }

    /* RX Data */
    while (ret == GQSPI_CODE_SUCCESS && rxData && rxSz > 0) {
        xferSz = rxSz;

        /* Enable RX */
        reg_genfifo &= ~(GQSPI_GEN_FIFO_TX | GQSPI_GEN_FIFO_IMM_MASK |
                         GQSPI_GEN_FIFO_EXP_MASK);
        reg_genfifo |= (GQSPI_GEN_FIFO_RX | GQSPI_GEN_FIFO_DATA_XFER);

        if (xferSz > GQSPI_GEN_FIFO_IMM_MASK) {
            /* Use exponent mode */
            xferSz = 256; /* 2 ^ 8 = 256 */
            reg_genfifo |= GQSPI_GEN_FIFO_EXP_MASK;
            reg_genfifo |= GQSPI_GEN_FIFO_IMM(8); /* IMM is exponent */
        }
        else {
            reg_genfifo |= GQSPI_GEN_FIFO_IMM(xferSz); /* IMM is length */
        }

        /* Submit general FIFO operation */
        ret = qspi_gen_fifo_write(reg_genfifo);
        if (ret != GQSPI_CODE_SUCCESS) {
            wolfBoot_printf("on line %d: error %d\n", __LINE__, ret);
            break;
        }

        /* Read FIFO */
        ret = gspi_fifo_rx(rxData, xferSz-dummySz, dummySz);
        if (ret != GQSPI_CODE_SUCCESS) {
            wolfBoot_printf("on line %d: error %d\n", __LINE__, ret);
        }

        /* offset size and buffer */
        rxSz -= xferSz;
        rxData += (xferSz - dummySz);
        dummySz = 0; /* only first RX */
    }

    qspi_cs(pDev, 0); /* Deselect Slave */
    GQSPI_EN = 0; /* Disable Device */

    return ret;
}
#endif /* QSPI Implementation */

static int qspi_flash_read_id(QspiDev_t* dev, uint8_t* id, uint32_t idSz)
{
    int ret;
    uint8_t cmd[20]; /* size multiple of uint32_t */
    uint8_t status = 0;

    memset(cmd, 0, sizeof(cmd));
    cmd[0] = MULTI_IO_READ_ID_CMD;
    ret = qspi_transfer(&mDev, cmd, 1, NULL, 0, cmd, sizeof(cmd), 0,
        GQSPI_GEN_FIFO_MODE_SPI);

    wolfBoot_printf("Read FlashID %s: Ret %d, %02x %02x %02x\n",
        (dev->cs & GQSPI_GEN_FIFO_CS_LOWER) ? "Lower" : "Upper",
        ret, cmd[0],  cmd[1],  cmd[2]);

    if (ret == GQSPI_CODE_SUCCESS && id) {
        if (idSz > sizeof(cmd))
            idSz = sizeof(cmd);
        memcpy(id, cmd, idSz);
    }

    qspi_status(dev, &status);
    if (status & WRITE_EN_MASK) {
        wolfBoot_printf("Write disabled: status %02x\n", status);
        ret = -1;
    }

    return ret;
}

static int qspi_write_enable(QspiDev_t* dev)
{
    int ret;
    uint8_t cmd[4]; /* size multiple of uint32_t */
    uint8_t status = 0;

    memset(cmd, 0, sizeof(cmd));
    cmd[0] = WRITE_ENABLE_CMD;
    ret = qspi_transfer(&mDev, cmd, 1, NULL, 0, NULL, 0, 0,
        GQSPI_GEN_FIFO_MODE_SPI);
#if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 2
    wolfBoot_printf("Write Enable: Ret %d\n", ret);
#endif
    ret = qspi_wait_ready(dev);
#if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 2
    wolfBoot_printf("Wait ready: Ret %d\n", ret);
#endif

    ret = qspi_wait_we(dev);
#if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 2
    wolfBoot_printf("Wait we: Ret %d\n", ret);
#endif

    qspi_status(dev, &status);
    if ((status & WRITE_EN_MASK) == 0) {
        wolfBoot_printf("Write enable failed: status %02x\n", status);
        ret = -1;
    }

    return ret;
}
static int qspi_write_disable(QspiDev_t* dev)
{
    int ret;
    uint8_t cmd[4]; /* size multiple of uint32_t */

    memset(cmd, 0, sizeof(cmd));
    cmd[0] = WRITE_DISABLE_CMD;
    ret = qspi_transfer(dev, cmd, 1, NULL, 0, NULL, 0, 0,
        GQSPI_GEN_FIFO_MODE_SPI);
#if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 2
    wolfBoot_printf("Write Disable: Ret %d\n", ret);
#endif
    return ret;
}

static int qspi_flash_status(QspiDev_t* dev, uint8_t* status)
{
    int ret;
    uint8_t cmd[4]; /* size multiple of uint32_t */

    memset(cmd, 0, sizeof(cmd));
    cmd[0] = READ_FSR_CMD;
    ret = qspi_transfer(dev, cmd, 1, NULL, 0, cmd, 2, 0,
        GQSPI_GEN_FIFO_MODE_SPI);
#if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 2
    wolfBoot_printf("Flash Status: Ret %d Cmd %02x %02x\n", ret, cmd[0], cmd[1]);
#endif
    if (ret == GQSPI_CODE_SUCCESS && status) {
        if (dev->stripe) {
            cmd[0] &= cmd[1];
        }
        *status = cmd[0];
    }
    return ret;
}

static int qspi_status(QspiDev_t* dev, uint8_t* status)
{
    int ret;
    uint8_t cmd[4]; /* size multiple of uint32_t */

    memset(cmd, 0, sizeof(cmd));
    cmd[0] = READ_SR_CMD;
    ret = qspi_transfer(dev, cmd, 1, NULL, 0, cmd, 2, 0,
        GQSPI_GEN_FIFO_MODE_SPI);
#if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 2
    wolfBoot_printf("Status: Ret %d Cmd %02x %02x\n", ret, cmd[0], cmd[1]);
#endif
    if (ret == GQSPI_CODE_SUCCESS && status) {
        if (dev->stripe) {
            cmd[0] &= cmd[1];
        }
        *status = cmd[0];
    }
    return ret;
}

static int qspi_wait_ready(QspiDev_t* dev)
{
    int ret;
    uint32_t timeout;
    uint8_t status = 0;

    timeout = 0;
    while (++timeout < QSPI_FLASH_READY_TRIES) {
        ret = qspi_flash_status(dev, &status);
        if (ret == GQSPI_CODE_SUCCESS && (status & FLASH_READY_MASK)) {
            return ret;
        }
    }

    wolfBoot_printf("Flash Ready Timeout!\n");

    return GQSPI_CODE_TIMEOUT;
}

static int qspi_wait_we(QspiDev_t* dev)
{
    int ret;
    uint32_t timeout;
    uint8_t status = 0;

    timeout = 0;
    while (++timeout < QSPI_FLASH_READY_TRIES) {
        ret = qspi_status(dev, &status);
        if (ret == GQSPI_CODE_SUCCESS &&
        (status & WRITE_EN_MASK)
        ) {
            return ret;
        }
    }

    wolfBoot_printf("Flash WE Timeout!\n");

    return GQSPI_CODE_TIMEOUT;
}


#if GQPI_USE_4BYTE_ADDR == 1
static int qspi_enter_4byte_addr(QspiDev_t* dev)
{
    int ret;
    uint8_t cmd[4]; /* size multiple of uint32_t */

    memset(cmd, 0, sizeof(cmd));
    cmd[0] = ENTER_4B_ADDR_MODE_CMD;
    (void)qspi_wait_ready(&mDev); /* Wait for not busy */
    ret = qspi_write_enable(&mDev);
    if (ret == GQSPI_CODE_SUCCESS) {
        ret = qspi_transfer(dev, cmd, 1, NULL, 0, NULL, 0, 0,
            GQSPI_GEN_FIFO_MODE_SPI);
    #if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 2
        wolfBoot_printf("Enter 4-byte address mode: Ret %d\n", ret);
    #endif
        if (ret == GQSPI_CODE_SUCCESS) {
            ret = qspi_wait_ready(&mDev); /* Wait for not busy */
        }
        qspi_write_disable(&mDev);
    }
    return ret;
}
static int qspi_exit_4byte_addr(QspiDev_t* dev)
{
    int ret;
    uint8_t cmd[4]; /* size multiple of uint32_t */

    memset(cmd, 0, sizeof(cmd));
    cmd[0] = EXIT_4B_ADDR_MODE_CMD;
    ret = qspi_write_enable(&mDev);
    if (ret == GQSPI_CODE_SUCCESS) {
        ret = qspi_transfer(dev, cmd, 1, NULL, 0, NULL, 0, 0,
            GQSPI_GEN_FIFO_MODE_SPI);
    #if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 2
        wolfBoot_printf("Exit 4-byte address mode: Ret %d\n", ret);
    #endif
        if (ret == GQSPI_CODE_SUCCESS) {
            ret = qspi_wait_ready(&mDev); /* Wait for not busy */
        }
        qspi_write_disable(&mDev);
    }
    return ret;
}
#endif

/* QSPI functions */
void qspi_init(uint32_t cpu_clock, uint32_t flash_freq)
{
    int ret;
    uint32_t reg_cfg, reg_isr;
    uint8_t id_low[4];
#if GQPI_USE_DUAL_PARALLEL == 1
    uint8_t id_hi[4];
#endif
    uint32_t timeout;
#ifdef USE_XQSPIPSU
    XQspiPsu_Config *QspiConfig;
#endif

    (void)cpu_clock;
    (void)flash_freq;

    memset(&mDev, 0, sizeof(mDev));

#ifdef USE_XQSPIPSU
    /* Xilinx BSP Driver */
    QspiConfig = XQspiPsu_LookupConfig(QSPI_DEVICE_ID);
    if (QspiConfig == NULL) {
        wolfBoot_printf("QSPI config lookup failed\n");
        return;
    }
    ret = XQspiPsu_CfgInitialize(&mDev.qspiPsuInst, QspiConfig, QspiConfig->BaseAddress);
    if (ret != 0) {
        wolfBoot_printf("QSPI config init failed\n");
        return;
    }
    XQspiPsu_SetOptions(&mDev.qspiPsuInst, XQSPIPSU_MANUAL_START_OPTION);
    XQspiPsu_SetClkPrescaler(&mDev.qspiPsuInst, QSPI_CLK_PRESACALE);

#elif defined(USE_QNX)
    /* QNX QSPI driver */
    mDev.qnx = xzynq_qspi_open();
    if (mDev.qnx == NULL) {
        wolfBoot_printf("QSPI failed to open\n");
        return;
    }
#else
    /* QSPI bare-metal driver */

    /* Disable Linear Mode in case FSBL enabled it */
    LQSPI_EN = 0;

    /* Select Generic Quad-SPI */
    GQSPI_SEL = 1;

    /* Clear and disable interrupts */
    reg_isr = GQSPI_ISR;
    GQSPI_ISR |= GQSPI_ISR_WR_TO_CLR_MASK; /* Clear poll timeout counter interrupt */
    reg_cfg = QSPIDMA_DST_I_STS;
    QSPIDMA_DST_I_STS = reg_cfg; /* clear all active interrupts */
    QSPIDMA_DST_STS |= QSPIDMA_DST_STS_WTC; /* mark outstanding DMA's done */
    GQSPI_IDR = GQSPI_IXR_ALL_MASK; /* disable interrupts */
    QSPIDMA_DST_I_STS = QSPIDMA_DST_I_STS_ALL_MASK; /* disable interrupts */
    /* Reset FIFOs */
    if (GQSPI_ISR & GQSPI_IXR_RX_FIFO_EMPTY) {
        GQSPI_FIFO_CTRL |= (GQSPI_FIFO_CTRL_RST_TX_FIFO | GQSPI_FIFO_CTRL_RST_RX_FIFO);
    }
    if (reg_isr & GQSPI_IXR_RX_FIFO_EMPTY) {
        GQSPI_FIFO_CTRL |= GQSPI_FIFO_CTRL_RST_RX_FIFO;
    }

    GQSPI_EN = 0; /* Disable device */

    /* Initialize clock divisor, write protect hold and start mode */
    reg_cfg  = GQSPI_CFG_MODE_EN_IO; /* Use I/O Transfer Mode */
    reg_cfg |= GQSPI_CFG_BAUD_RATE_DIV(GQSPI_CLK_DIV); /* Clock Divider */
    reg_cfg |= GQSPI_CFG_WP_HOLD; /* Use WP Hold */
    reg_cfg |= GQSPI_CFG_START_GEN_FIFO; /* Start GFIFO command execution */
    reg_cfg &= ~(GQSPI_CFG_CLK_POL | GQSPI_CFG_CLK_PH); /* Use POL=0,PH=0 */
    GQSPI_CFG = reg_cfg;

#if GQSPI_CLK_DIV >= 2 /* 300/8=37.5MHz */
    /* At 40 MHz, the Quad-SPI controller should be in non-loopback mode with
     * the clock and data tap delays bypassed. */
    IOU_TAPDLY_BYPASS |= IOU_TAPDLY_BYPASS_LQSPI_RX;
    GQSPI_LPBK_DLY_ADJ = 0;
    QSPI_DATA_DLY_ADJ = 0;
#elif GQSPI_CLK_DIV >= 1 /* 300/4=75MHz */
    /* At 100 MHz, the Quad-SPI controller should be in clock loopback mode
     * with the clock tap delay bypassed, but the data tap delay enabled. */
    IOU_TAPDLY_BYPASS |= IOU_TAPDLY_BYPASS_LQSPI_RX;
    GQSPI_LPBK_DLY_ADJ = GQSPI_LPBK_DLY_ADJ_USE_LPBK;
    QSPI_DATA_DLY_ADJ = QSPI_DATA_DLY_ADJ_USE_DATA_DLY | QSPI_DATA_DLY_ADJ_DATA_DLY_ADJ(2);
#else
    /* At 150 MHz, only the generic controller can be used.
     * The generic controller should be in clock loopback mode and the clock
     * tap delay enabled, but the data tap delay disabled. */
    IOU_TAPDLY_BYPASS = 0;
    GQSPI_LPBK_DLY_ADJ = GQSPI_LPBK_DLY_ADJ_USE_LPBK;
    QSPI_DATA_DLY_ADJ = 0;
#endif

    /* Initialize hardware parameters for Threshold and Interrupts */
    GQSPI_TX_THRESH = 1;
    GQSPI_RX_THRESH = 1;
    GQSPI_GF_THRESH = 16;

    /* Reset DMA */
    QSPIDMA_DST_CTRL = QSPIDMA_DST_CTRL_DEF;
    QSPIDMA_DST_CTRL2 = QSPIDMA_DST_CTRL2_DEF;

    /* Interrupts unmask and enable */
    GQSPI_IMR = GQSPI_IXR_ALL_MASK;
    GQSPI_IER = GQSPI_IXR_ALL_MASK;

    GQSPI_EN = 1; /* Enable Device */
#endif /* USE_QNX */
    (void)reg_cfg;
    (void)reg_isr;

    /* ------ Flash Read ID (retry) ------ */
    timeout = 0;
    while (++timeout < QSPI_FLASH_READY_TRIES) {
        /* Slave Select - lower chip */
        mDev.mode = GQSPI_GEN_FIFO_MODE_SPI;
        mDev.bus = GQSPI_GEN_FIFO_BUS_LOW;
        mDev.cs = GQSPI_GEN_FIFO_CS_LOWER;
        ret = qspi_flash_read_id(&mDev, id_low, sizeof(id_low));
        if (ret != GQSPI_CODE_SUCCESS) {
            continue;
        }

    #if GQPI_USE_DUAL_PARALLEL == 1
        /* Slave Select - upper chip */
        mDev.mode = GQSPI_GEN_FIFO_MODE_SPI;
        mDev.bus = GQSPI_GEN_FIFO_BUS_UP;
        mDev.cs = GQSPI_GEN_FIFO_CS_UPPER;
        ret = qspi_flash_read_id(&mDev, id_hi, sizeof(id_hi));
        if (ret != GQSPI_CODE_SUCCESS) {
            continue;
        }

        /* ID's for upper and lower must match */
        if ((id_hi[0] == 0 || id_hi[0] == 0xFF) ||
            (id_hi[0] != id_low[0] &&
            id_hi[1] != id_low[1] &&
            id_hi[2] != id_low[2]))
        {
            wolfBoot_printf("Flash ID error!\n");
            continue;
        }
    #endif
        break; /* success */
    }

    /* Slave Select */
    mDev.mode = GQSPI_QSPI_MODE;
#if GQPI_USE_DUAL_PARALLEL == 1
    mDev.bus = GQSPI_GEN_FIFO_BUS_BOTH;
    mDev.cs = GQSPI_GEN_FIFO_CS_BOTH;
    mDev.stripe = GQSPI_GEN_FIFO_STRIPE;
#endif

#if GQPI_USE_4BYTE_ADDR == 1
    /* Enter 4-byte address mode */
    ret = qspi_enter_4byte_addr(&mDev);
    if (ret != GQSPI_CODE_SUCCESS)
        return;
#endif

#ifdef TEST_EXT_FLASH
    test_ext_flash(&mDev);
#endif
}

#if 0
uint64_t hal_timer_ms(void)
{
    uint64_t val;
    unsigned long cntfrq;
    unsigned long cntpct;
    asm volatile("mrs %0, cntfrq_el0" : "=r" (cntfrq));
    asm volatile("mrs %0, cntpct_el0" : "=r" (cntpct));
    val = cntpct * 1000;
    val /= cntfrq;
    return val;
}
#endif

/* public HAL functions */
void hal_init(void)
{
    uint32_t cpu_freq = 0;
    const char* bootMsg = "\nwolfBoot Secure Boot\n";

#ifdef DEBUG_UART
    uart_init();
#endif
    wolfBoot_printf(bootMsg);

#if 0
    /* This is only allowed for EL-3 */
    asm volatile("msr cntfrq_el0, %0" : : "r" (cpu_freq) : "memory");
#endif

    qspi_init(cpu_freq, 0);
}

void hal_prepare_boot(void)
{
#if GQPI_USE_4BYTE_ADDR == 1
    /* Exit 4-byte address mode */
    int ret = qspi_exit_4byte_addr(&mDev);
    if (ret != GQSPI_CODE_SUCCESS)
        return;
#endif

#ifdef USE_QNX
    if (mDev.qnx) {
        xzynq_qspi_close(mDev.qnx);
        mDev.qnx = NULL;
    }
#endif
}

/* Flash functions must be relocated to RAM for execution */
int RAMFUNCTION hal_flash_write(uintptr_t address, const uint8_t *data, int len)
{
    return 0;
}

void RAMFUNCTION hal_flash_unlock(void)
{
}

void RAMFUNCTION hal_flash_lock(void)
{
}


int RAMFUNCTION hal_flash_erase(uintptr_t address, int len)
{
    return 0;
}

/* Xilinx Write uses SPI mode and Page Program 0x02 */
/* Issues using write with QSPI mode */
int RAMFUNCTION ext_flash_write(uintptr_t address, const uint8_t *data, int len)
{
    int ret = 0;
    uint8_t cmd[8]; /* size multiple of uint32_t */
    uint32_t xferSz, page, pages, idx;
    uintptr_t addr;

#if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 2
    wolfBoot_printf("Flash Write: Addr 0x%x, Ptr %p, Len %d\n",
        address, data, len);
#endif

    /* write by page */
    pages = ((len + (FLASH_PAGE_SIZE-1)) / FLASH_PAGE_SIZE);
    for (page = 0; page < pages; page++) {
        ret = qspi_write_enable(&mDev);
        if (ret != GQSPI_CODE_SUCCESS) {
            break;
        }
        xferSz = len;
        if (xferSz > FLASH_PAGE_SIZE)
            xferSz = FLASH_PAGE_SIZE;

        addr = address + (page * FLASH_PAGE_SIZE);
        if (mDev.stripe) {
            /* For dual parallel the address divide by 2 */
            addr /= 2;
        }

        /* ------ Write Flash (page at a time) ------ */
        memset(cmd, 0, sizeof(cmd));
        idx = 0;
        cmd[idx++] = PAGE_PROG_CMD;
    #if GQPI_USE_4BYTE_ADDR == 1
        cmd[idx++] = ((addr >> 24) & 0xFF);
    #endif
        cmd[idx++] = ((addr >> 16) & 0xFF);
        cmd[idx++] = ((addr >> 8)  & 0xFF);
        cmd[idx++] = ((addr >> 0)  & 0xFF);
        ret = qspi_transfer(&mDev, cmd, idx,
            (const uint8_t*)(data + (page * FLASH_PAGE_SIZE)),
            xferSz, NULL, 0, 0, GQSPI_GEN_FIFO_MODE_SPI);
        wolfBoot_printf("Flash Page %d Write: Ret %d\n", page, ret);
        if (ret != GQSPI_CODE_SUCCESS)
            break;

        ret = qspi_wait_ready(&mDev); /* Wait for not busy */
        if (ret != GQSPI_CODE_SUCCESS) {
            break;
        }
        qspi_write_disable(&mDev);
        len -= xferSz;
    }

    return ret;
}

#if GQSPI_QSPI_MODE == GQSPI_GEN_FIFO_MODE_QSPI && GQPI_USE_4BYTE_ADDR == 1
#define FLASH_READ_CMD QUAD_READ_4B_CMD
#elif GQSPI_QSPI_MODE == GQSPI_GEN_FIFO_MODE_DSPI && GQPI_USE_4BYTE_ADDR == 1
#define FLASH_READ_CMD DUAL_READ_4B_CMD
#elif GQPI_USE_4BYTE_ADDR == 1
#define FLASH_READ_CMD FAST_READ_4B_CMD
#elif GQSPI_QSPI_MODE == GQSPI_GEN_FIFO_MODE_QSPI
#define FLASH_READ_CMD QUAD_READ_CMD
#elif GQSPI_QSPI_MODE == GQSPI_GEN_FIFO_MODE_DSPI
#define FLASH_READ_CMD DUAL_READ_CMD
#else
#define FLASH_READ_CMD FAST_READ_CMD
#endif

int RAMFUNCTION ext_flash_read(uintptr_t address, uint8_t *data, int len)
{
    int ret;
    uint8_t cmd[8]; /* size multiple of uint32_t */
    uint32_t idx = 0;

#if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 2
    wolfBoot_printf("Flash Read: Addr 0x%x, Ptr %p, Len %d\n",
        address, data, len);
#endif

    if (mDev.stripe) {
        /* For dual parallel the address divide by 2 */
        address /= 2;
    }

    /* ------ Read Flash ------ */
    memset(cmd, 0, sizeof(cmd));
    cmd[idx++] = FLASH_READ_CMD;
#if GQPI_USE_4BYTE_ADDR == 1
    cmd[idx++] = ((address >> 24) & 0xFF);
#endif
    cmd[idx++] = ((address >> 16) & 0xFF);
    cmd[idx++] = ((address >> 8)  & 0xFF);
    cmd[idx++] = ((address >> 0)  & 0xFF);
    ret = qspi_transfer(&mDev, cmd, idx, NULL, 0, data, len, GQSPI_DUMMY_READ,
        mDev.mode);
#if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 2
    wolfBoot_printf("Flash Read: Ret %d\r\n", ret);
#endif

    return (ret == 0) ? len : ret;
}

/* Issues a sector erase based on flash address */
int RAMFUNCTION ext_flash_erase(uintptr_t address, int len)
{
    int ret = 0;
    uint8_t cmd[8]; /* size multiple of uint32_t */
    uint32_t idx = 0;
    uintptr_t qspiaddr;

#if defined(DEBUG_ZYNQ) && DEBUG_ZYNQ >= 2
    wolfBoot_printf("Flash Erase: Addr 0x%x, Len %d\n",  address, len);
#endif

    while (len > 0) {
        /* For dual parallel the address divide by 2 */
        qspiaddr = (mDev.stripe) ? address / 2 : address;

        ret = qspi_write_enable(&mDev);
        if (ret == GQSPI_CODE_SUCCESS) {
            /* ------ Erase Flash ------ */
            memset(cmd, 0, sizeof(cmd));
            cmd[idx++] = SEC_ERASE_CMD;
        #if GQPI_USE_4BYTE_ADDR == 1
            cmd[idx++] = ((qspiaddr >> 24) & 0xFF);
        #endif
            cmd[idx++] = ((qspiaddr >> 16) & 0xFF);
            cmd[idx++] = ((qspiaddr >> 8)  & 0xFF);
            cmd[idx++] = ((qspiaddr >> 0)  & 0xFF);
            ret = qspi_transfer(&mDev, cmd, idx, NULL, 0, NULL, 0, 0,
                GQSPI_GEN_FIFO_MODE_SPI);
            wolfBoot_printf("Flash Erase: Ret %d\n", ret);
            if (ret == GQSPI_CODE_SUCCESS) {
                ret = qspi_wait_ready(&mDev); /* Wait for not busy */
            }
            qspi_write_disable(&mDev);
        }

        address += WOLFBOOT_SECTOR_SIZE;
        len -= WOLFBOOT_SECTOR_SIZE;
    }

    return ret;
}

void RAMFUNCTION ext_flash_lock(void)
{

}

void RAMFUNCTION ext_flash_unlock(void)
{

}

#ifdef MMU
void* hal_get_dts_address(void)
{
    return (void*)WOLFBOOT_DTS_BOOT_ADDRESS;
}

int hal_dts_fixup(void* dts_addr)
{
    /* place FDT fixup specific to ZynqMP here */
    //fdt_set_boot_cpuid_phys(buf, fdt_boot_cpuid_phys(fdt));
    return 0;
}
#endif


#ifdef TEST_EXT_FLASH
#ifndef TEST_EXT_ADDRESS
#define TEST_EXT_ADDRESS 0x2800000 /* 40MB */
#endif
static int test_ext_flash(QspiDev_t* dev)
{
    int ret;
    uint32_t i;
    uint8_t pageData[FLASH_PAGE_SIZE*4];

#ifndef TEST_FLASH_READONLY
    /* Erase sector */
    ret = ext_flash_erase(TEST_EXT_ADDRESS, WOLFBOOT_SECTOR_SIZE);
    wolfBoot_printf("Erase Sector: Ret %d\n", ret);

    /* Write Pages */
    for (i=0; i<sizeof(pageData); i++) {
        pageData[i] = (i & 0xff);
    }
    ret = ext_flash_write(TEST_EXT_ADDRESS, pageData, sizeof(pageData));
    wolfBoot_printf("Write Page: Ret %d\n", ret);
#endif /* !TEST_FLASH_READONLY */

    /* Read page */
    memset(pageData, 0, sizeof(pageData));
    ret = ext_flash_read(TEST_EXT_ADDRESS, pageData, sizeof(pageData));
    wolfBoot_printf("Read Page: Ret %d\n", ret);

    wolfBoot_printf("Checking...\n");
    /* Check data */
    for (i=0; i<sizeof(pageData); i++) {
        wolfBoot_printf("check[%3d] %02x\n", i, pageData[i]);
        if (pageData[i] != (i & 0xff)) {
            wolfBoot_printf("Check Data @ %d failed\n", i);
            return GQSPI_CODE_FAILED;
        }
    }

    wolfBoot_printf("Flash Test Passed\n");
    return ret;
}
#endif /* TEST_EXT_FLASH */

#endif /* TARGET_zynq */
